--[[
    FactoryLib
    Copyright (C) 2008 Gael Lalire
    Contact: gael.lalire@laposte.net

    FactoryLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FactoryLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FactoryLib.  If not, see <http://www.gnu.org/licenses/>.
--]]

factoryLib={}
local factoryLib = LibStub:NewLibrary("FactoryLib-1.0", 1);
if not factoryLib then
  return;
end

local Factory = {};
factoryLib.Factory = Factory;

-- returns type, singleton
function Factory.CreateSingletonType(superFactoryType, ...)
  local instanceType, _, factory = Factory.CreateType(superFactoryType);
  local singleton = factory:Create(...);
  return instanceType, singleton;
end

-- returns type, factoryType
function Factory.CreateAbstractType(superFactoryType)
  local instanceType, factoryType = Factory.CreateType(superFactoryType);
  return instanceType, factoryType;
end

-- returns type, factoryType, factorySingleton
function Factory.CreateType(superFactoryType)
  local instanceType = {};
  local factoryType = {};
  setmetatable(factoryType, {__index=Factory});  
  if superFactoryType ~= nil then
    setmetatable(instanceType, {__index=superFactoryType.instanceType});
    factoryType.superFactoryType = superFactoryType;
  end
  factoryType.instanceType = instanceType;
  
  -- create the factory
  local factory = {};
  factory.factoryType = factoryType;
  factory.currentFactoryType = factoryType;
  setmetatable(factory, {__index = factoryType, __newindex=function (_, n) 
    error("A factory should not change its variables ("..n..")");
  end});
  return instanceType, factoryType, factory;
end

function Factory:DoInit(instance, ...)
  local initMethod = rawget(self.currentFactoryType.instanceType, "Init");
  initMethod(instance, ...);
end

-- this method should be overrided
function Factory:Create(...)
  local instance = self:NewInstance(...);
  local initMethod = rawget(self.currentFactoryType.instanceType, "Init");
  if initMethod ~= nil then
    initMethod(instance);
  end
  return instance;
end

function Factory:NewInstance(...)
  local factoryType = self.currentFactoryType;
  local superFactoryType = factoryType.superFactoryType;
  if superFactoryType == nil then
    local instance = {};
    setmetatable(instance, {__index=self.instanceType});
    return instance;
  else
    rawset(self, "currentFactoryType", superFactoryType);
    local instance = superFactoryType.Create(self, ...);
    rawset(self, "currentFactoryType", factoryType);
    return instance;
  end
end

function Factory:Cast(instance)
  setmetatable(instance, {__index=self.instanceType});
end

setmetatable(Factory, {__newindex=function () 
  error("Factory should not be modified");
end});

