--[[
    AbstractActionBarRendererLib
    Copyright (C) 2008 Gael Lalire
    Contact: gael.lalire@laposte.net

    AbstractActionBarRendererLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    AbstractActionBarRendererLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AbstractActionBarRendererLib.  If not, see <http://www.gnu.org/licenses/>.
--]]

local abstractActionBarRendererLib = LibStub:NewLibrary("AbstractActionBarRendererLib-1.0", 1);
if not abstractActionBarRendererLib then
  return;
end

-- begin import
local factoryLib = LibStub("FactoryLib-1.0");
local Factory = factoryLib.Factory;
local actionBarLib = LibStub("ActionBarLib-1.0");
local ActionBarRendererFactory = actionBarLib.ActionBarRendererFactory;
local actionLib = LibStub("ActionLib-1.0");
local Action = actionLib.Action;
local environmentLib = LibStub("EnvironmentLib-1.0");
local environment = environmentLib.environment;
-- end import

--[[
  The AbstractActionBarRenderer class.
--]]

-- AbstractActionBarRenderer extends ActionBarRenderer
local AbstractActionBarRenderer, AbstractActionBarRendererFactory = Factory.CreateAbstractType(ActionBarRendererFactory);
abstractActionBarRendererLib.AbstractActionBarRenderer = AbstractActionBarRenderer;
abstractActionBarRendererLib.AbstractActionBarRendererFactory = AbstractActionBarRendererFactory;

function AbstractActionBarRenderer:SPELL_UPDATE_COOLDOWN()
  local secureUnitButton = self.selectedSecureUnitButton;
  if secureUnitButton ~= nil then
    local action = secureUnitButton.action;
    if action ~= nil and action:IsAvailable() then
      local start, duration = action:GetCooldown();
      CooldownFrame_SetTimer(secureUnitButton.cooldown, start, duration, 1);
    end
  end
  for i=1, #self.secureUnitButton do
    secureUnitButton = self.secureUnitButton[i];  
    if secureUnitButton ~= nil then
      action = secureUnitButton.action;
      if action ~= nil and action:IsAvailable() then
        start, duration = action:GetCooldown();
        CooldownFrame_SetTimer(secureUnitButton.cooldown, start, duration, 1);
      end
    end
  end
end

function AbstractActionBarRenderer:ApplySkin(secureUnitButton)
  local position = secureUnitButton.position;
  
  local borderSize = self.borderSize;
  local borderTexture = secureUnitButton.borderTexture;
  borderTexture:SetPoint("TOPLEFT", secureUnitButton, -borderSize, borderSize);
  borderTexture:SetPoint("BOTTOMRIGHT", secureUnitButton, borderSize, -borderSize);
  if position == 0 then
    self.selectedSecureUnitButton = secureUnitButton; -- todo : add a lot of information here
    borderTexture:SetTexture(self.selectedTexture);  
  else
    self.secureUnitButton[position] = secureUnitButton; -- todo : add a lot of information here ... { ... , borderTexture, targetText, ... }
    borderTexture:SetTexture(self.texture);  
  end
end

function AbstractActionBarRenderer:UpdateTooltip(secureUnitButton)
  GameTooltip:SetOwner(secureUnitButton, "ANCHOR_RIGHT");
  secureUnitButton.action:ShowInTooltip(GameTooltip);
  if self.tooltipHelpAdded then
    GameTooltip:AddLine(" ");
    if secureUnitButton.position == 0 then
      self:AddSelectedTooltipHelp(GameTooltip);
    else
      if self.selectedSecureUnitButton:IsVisible() then
        self:AddFavoriteTooltipHelp(GameTooltip);
      else
        self:AddOpenTooltipHelp(GameTooltip);
      end
    end
  end
  GameTooltip:Show();
end

function AbstractActionBarRenderer:UnApplySkin(secureUnitButton)
  -- hide all graphical adds to the button
  local position = secureUnitButton.position;
  if position == 0 then
    self.selectedSecureUnitButton = nil;
  else
    self.secureUnitButton[position] = nil;
  end
end


function AbstractActionBarRenderer:ActionChanged(secureUnitButton)
  local action = secureUnitButton.action;
  
  if action ~= nil and action:IsAvailable() then
    local start, duration = action:GetCooldown();
    CooldownFrame_SetTimer(secureUnitButton.cooldown, start, duration, 1);
  
    if GameTooltip:GetOwner()==secureUnitButton then
      self:UpdateTooltip(secureUnitButton);
    end
    local texture = action:GetTexture();
    secureUnitButton:SetNormalTexture(texture);
    
    local target = action:GetTarget();
    if target=="target" then
      secureUnitButton.targetText:SetText("");
    elseif target=="player" then
      secureUnitButton.targetText:SetText("P");    
    elseif target=="focus" then
      secureUnitButton.targetText:SetText("F");
    elseif target=="targettarget" then
      secureUnitButton.targetText:SetText("A");  
    elseif target=="mouseover" then
      secureUnitButton.targetText:SetText("M");  
    elseif target==Action.MAIN_HAND_WEAPON then
      secureUnitButton.targetText:SetText("R");
    elseif target==Action.OFF_HAND_WEAPON then
      secureUnitButton.targetText:SetText("L");    
    else
      secureUnitButton.targetText:SetText("?");     
    end
  end
end

function AbstractActionBarRenderer:UpdateAllSkin()
  if self.selectedSecureUnitButton ~= nil then
    self:ApplySkin(self.selectedSecureUnitButton);
  end
  for _, secureUnitButton in ipairs(self.secureUnitButton) do
    self:ApplySkin(secureUnitButton);
  end
end

--[[
  The AbstractActionBarRendererFactory class.
--]]

-- left and top indicate the position of the first button's center
function AbstractActionBarRendererFactory:Create(tooltipLabels, texture, selectedTexture, borderSize)
  local abstractActionBarRenderer = self:NewInstance(tooltipLabels);
  
  environment:RegisterEvent("SPELL_UPDATE_COOLDOWN", abstractActionBarRenderer);

  abstractActionBarRenderer.selectedSecureUnitButton = nil;
  abstractActionBarRenderer.secureUnitButton = {};
  abstractActionBarRenderer.texture = texture;
  abstractActionBarRenderer.selectedTexture = selectedTexture;
  abstractActionBarRenderer.borderSize = borderSize;
  
  return abstractActionBarRenderer;
end
