--[[
    CircleActionBarRendererLib
    Copyright (C) 2008 Gael Lalire
    Contact: gael.lalire@laposte.net

    CircleActionBarRendererLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CircleActionBarRendererLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CircleActionBarRendererLib.  If not, see <http://www.gnu.org/licenses/>.
--]]

local circleActionBarRendererLib = LibStub:NewLibrary("CircleActionBarRendererLib-1.0", 1);
if not circleActionBarRendererLib then
  return;
end

-- begin import
local factoryLib = LibStub("FactoryLib-1.0");
local Factory = factoryLib.Factory;
local abstractActionBarRendererLib = LibStub("AbstractActionBarRendererLib-1.0");
local AbstractActionBarRendererFactory = abstractActionBarRendererLib.AbstractActionBarRendererFactory;
local AbstractActionBarRenderer = abstractActionBarRendererLib.AbstractActionBarRenderer;
local environmentLib = LibStub("EnvironmentLib-1.0");
local environment = environmentLib.environment;
-- end import

--[[
  The CircleActionBarRenderer class.
--]]

-- CircleActionBarRenderer extends ActionBarRenderer
local CircleActionBarRenderer, CircleActionBarRendererFactory, circleActionBarRendererFactory = Factory.CreateType(AbstractActionBarRendererFactory);
circleActionBarRendererLib.CircleActionBarRenderer = CircleActionBarRenderer;
circleActionBarRendererLib.CircleActionBarRendererFactory = CircleActionBarRendererFactory;
circleActionBarRendererLib.circleActionBarRendererFactory = circleActionBarRendererFactory;


function CircleActionBarRenderer:ApplySkin(secureUnitButton)
  local position = secureUnitButton.position;

  -- Show all graphical adds to the button
  local buttonSize = self.buttonSide;

  secureUnitButton:SetWidth(buttonSize);
  secureUnitButton:SetHeight(buttonSize);
  secureUnitButton:SetPoint("CENTER", self.point, "TOPLEFT", self.left+position*(self.buttonDiameter*math.cos(self.angle)), -(self.top-position*(self.buttonDiameter*math.sin(self.angle))));

  -- call parent
  AbstractActionBarRenderer.ApplySkin(self, secureUnitButton);  
end

function CircleActionBarRenderer:UnApplySkin(secureUnitButton)
  -- call parent
  AbstractActionBarRenderer.UnApplySkin(self, secureUnitButton);
end


function CircleActionBarRenderer:SetAngle(angle)
  self.angle = angle;
  self:UpdateAllSkin();
end

function CircleActionBarRenderer:SetPosition(point, left, top)
  self.point = point;
  self.left = left;
  self.top = top;
  self:UpdateAllSkin();  
end

--[[
  The CircleActionBarRendererFactory class.
--]]

-- left and top indicate the position of the first button's center
function CircleActionBarRendererFactory:Create(tooltipLabels, buttonDiameter, texture, selectedTexture)
  
  local buttonSide = buttonDiameter / 1.4142; -- sqrt(2)  

  local circleActionBarRenderer = self:NewInstance(tooltipLabels, texture, selectedTexture, buttonSide*.3);
  circleActionBarRenderer.buttonSide = buttonSide;
  circleActionBarRenderer.point = UIParent;
  circleActionBarRenderer.left = 0;
  circleActionBarRenderer.top = 0;
  circleActionBarRenderer.angle = 0;
  circleActionBarRenderer.buttonDiameter = buttonDiameter;
  
  return circleActionBarRenderer;
end
