--[[
    ManagerFrameLib
    Copyright (C) 2008 Gael Lalire
    Contact: gael.lalire@laposte.net

    ManagerFrameLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    ManagerFrameLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ManagerFrameLib.  If not, see <http://www.gnu.org/licenses/>.
--]]

local managerFrameLib = LibStub:NewLibrary("ManagerFrameLib-1.0", 1);
if not managerFrameLib then
  return;
end

-- begin import
local factoryLib = LibStub("FactoryLib-1.0");
local Factory = factoryLib.Factory;
local environmentLib = LibStub("EnvironmentLib-1.0");
local environment = environmentLib.environment;
local tM2TimerRendererLib = LibStub("TM2TimerRendererLib-1.0");
local libSinkTimerRendererLib = LibStub("LibSinkTimerRendererLib-1.0");
local actionLib = LibStub("ActionLib-1.0");
local actionBarLib = LibStub("ActionBarLib-1.0");
local timerLib = LibStub("TimerLib-1.0");
-- end import

local function rankToRoman(rank)
  if rank > 18 then
    error("Unsupported rank");
    return;
  end
  local m5 = rank%5;
  local before, after = "", "";
  if m5 == 1 then
    after = "I";  
  elseif m5 == 2 then
    after = "II";
  elseif m5 == 3 then
    after = "III";
  elseif m5 == 4 then
    before = "I";
  end
  local rest = math.floor((rank + 1) / 5);
  if rest == 0 then
    return before..after; -- 0 to 3
  elseif rest == 1 then
    return before.."V"..after; -- 4 to 8
  elseif rest == 2 then
    return before.."X"..after; -- 9 to 13
  elseif rest == 3 then
    return "X"..before.."V"..after;  -- 14 to 18
  end  
end

local function romanToRank(roman)
  local total = 0;
  local base = 0;
  
  local pos = string.len(roman);
  
  while pos >= 1 do
    local c = string.sub(roman, pos, pos);

    if c == "X" then
      if base <= 10 then
        total = total + 10;
        base = 10;
      else
        total = total - 10;
      end
    elseif c == "V" then
      if base <= 5 then
        total = total + 5;
        base = 5;  
      else
        total = total - 5;
      end
    elseif c == "I" then
      if base <= 1 then
        total = total + 1;
        base = 1;
      else
        total = total - 1;
      end
    else
      error("Unsupported rank");
     return;
    end
    pos = pos - 1;
  end
  return total;
end


local function transformSummonArg(timer, initialSpellArg, action, rank, targetName)
  -- after stop timer we can change updater safely
  local res = {};
  
  local page = nil;
  
  if action.spellName == totemManager.FIRST_CALL then
    page = 1;
  elseif action.spellName == totemManager.SECOND_CALL then
    page = 2;
  elseif action.spellName == totemManager.THIRD_CALL then
    page = 3;
  else
    return;
  end

  if page ~= nil then

    environment:RegisterEvent("PLAYER_TOTEM_UPDATE", {PLAYER_TOTEM_UPDATE=function (tself)
      local haveTotem, name, startTime, duration, icon = GetTotemInfo(tself.slot);
      
      if not haveTotem then
        -- this element as no change we can stop watching
        environment:UnregisterEvent("PLAYER_TOTEM_UPDATE", tself);
        return;
      end
      
      local tdiff = math.abs(tself.castTime - GetTime());
      if tdiff >= 3 then
        -- after 3 seconds no watch
        environment:UnregisterEvent("PLAYER_TOTEM_UPDATE", tself);  
        return;    
      end

      if name==nil or startTime==0 then
          -- this element seems unchanged (but we wait 3 seconds)      
        return;
      end
      
      local tdiffStart = math.abs(tself.castTime - startTime);
      if tdiffStart >= 1 then
          -- this element seems unchanged (but we wait 3 seconds)      
        return;
      end      

      local _, _, totemName, latinTotemRank = string.find(name, "(.*) ([XVI]+)");
      if totemName == nil then
        totemName = name; -- unranked totem
      end
      local totemRank = nil;
      if latinTotemRank ~= nil then
        totemRank = romanToRank(latinTotemRank);
      end 

      local nInitialSpellArg = tself.propertiesByTotemName[string.lower(totemName)];
      if nInitialSpellArg == nil then
        for _, nISA in pairs(tself.propertiesByTotemName) do
          if nISA.action:GetTexture() == icon then
            nInitialSpellArg = nISA;
            break;
          end
        end
      end
      if nInitialSpellArg == nil then
        return;
      end
        
      if tself.lastName == name then
        -- update is not useful (duration is updated before name)
        return;
      end
                  
      timer:StopTimer();
      tself.lastName = name;
      res.name = name;
      res.target = nil;
      if nInitialSpellArg.duration == nil then
        res.duration = 300; -- default to 5 minutes
      elseif type(nInitialSpellArg.duration)=="function" then
        res.duration = nInitialSpellArg.duration(totemRank);
      else
        res.duration = nInitialSpellArg.duration;
      end
      if nInitialSpellArg.totemName ~= nil then
        res.duration = ((5 + environment:GetTalentRankByIndex(5, 3)) * res.duration) /5;
      end
      res.texture = icon;
      if nInitialSpellArg.tickFrequency ~= nil then
        res.tickFrequency = nInitialSpellArg.tickFrequency;
        res.tickStart = nInitialSpellArg.tickStart or 0.5;
      end
      res.updater = nInitialSpellArg.updater;
      if res.updater ~= nil then
        if type(nInitialSpellArg.raidEffect)=="function" then
          res.updater.raidEffect = nInitialSpellArg.raidEffect();
        else
          res.updater.raidEffect = nInitialSpellArg.raidEffect;
        end
        if type(nInitialSpellArg.effect)=="function" then
          res.updater.effect = nInitialSpellArg.effect();
        else
         res.updater.effect = nInitialSpellArg.effect;    
        end
        res.updater.found = false;
      end
      timer:StartTimer(res.name, res.duration, res.texture, res.target, res.count,
          res.tickFrequency, res.tickStart, res.updater);
      timer:SetTimeLeft(res.duration - GetTime() + tself.castTime);          
    end, slot=initialSpellArg.slot, propertiesByTotemName=initialSpellArg.propertiesByTotemName, castTime=GetTime()});
  end
  
end


-- this function is necessary to spell action
local function transformSpellArg(timer, initialSpellArg, action, rank, targetName)
  -- after stop timer we can change updater safely
  local res = {};    
  
  timer:StopTimer();
  if initialSpellArg.totemName and rank ~= -1 and rank ~= 1 then
    res.name = initialSpellArg.totemName.." "..rankToRoman(rank);
  else
    res.name = initialSpellArg.totemName or initialSpellArg.buffName or initialSpellArg.name;
  end
  res.target = targetName;
  if initialSpellArg.duration == nil then
    res.duration = 300; -- default to 5 minutes
  elseif type(initialSpellArg.duration)=="function" then
    res.duration = initialSpellArg.duration(rank);
  else
    res.duration = initialSpellArg.duration;
  end
  if initialSpellArg.totemName ~= nil then
    res.duration = ((5 + environment:GetTalentRankByIndex(5, 3)) * res.duration) /5;
  end
  res.texture = action:GetTexture();
  if initialSpellArg.tickFrequency ~= nil then
    res.tickFrequency = initialSpellArg.tickFrequency;
    res.tickStart = initialSpellArg.tickStart or 0.5;
  end
  res.updater = initialSpellArg.updater;
  if res.updater ~= nil then
    if type(initialSpellArg.raidEffect)=="function" then
      res.updater.raidEffect = initialSpellArg.raidEffect();
    else
      res.updater.raidEffect = initialSpellArg.raidEffect;
    end
    if type(initialSpellArg.effect)=="function" then
      res.updater.effect = initialSpellArg.effect();
    else
     res.updater.effect = initialSpellArg.effect;    
    end
    res.updater.found = false;
  end
  timer:StartTimer(res.name, res.duration, res.texture, res.target, res.count,
      res.tickFrequency, res.tickStart, res.updater);
end

local function earthShieldTransformSpellArg(timer, initialSpellArg, action, _, targetName)
  local res = {};
  if UnitName("player")~=targetName then
    -- after stop timer we can change updater safely
    timer:StopTimer();
    initialSpellArg.updater.found = false;
    initialSpellArg.updater.targetName = targetName;
    initialSpellArg.updater.targetPath = environment:FindPath(targetName);
    timer:StartTimer(initialSpellArg.name, initialSpellArg.duration, action:GetTexture(), targetName, nil,
      nil, nil, initialSpellArg.updater);
  end
end

local function transformItemArg(timer, initialSpellArg, action, targetName)
  -- after stop timer we can change updater safely
  timer:StopTimer();
  local res = {};
  res.name = initialSpellArg.totemName or initialSpellArg.name;
  if initialSpellArg.duration == nil then
    res.duration = 120; -- default to 2 minutes
  else
    res.duration = initialSpellArg.duration;
  end
  res.texture = action:GetTexture();
  res.updater = initialSpellArg.updater;
  if res.updater ~= nil then
     res.updater.isRaidEffect = initialSpellArg.isRaidEffect;
     res.updater.effect = initialSpellArg.effect;
     res.updater.found = false;
  end
  timer:StartTimer(res.name, res.duration, res.texture, res.target, res.count,
      res.tickFrequency, res.tickStart, res.updater);
end

local function transformCooldownArg(timer, initialSpellArg, action)
  if action:IsAvailable() then
    -- after stop timer we can change updater safely
    timer:StopTimer();
    local res = {};
    res.name = initialSpellArg.name;
    local start, duration = action:GetCooldown();

    local time = GetTime();
    -- verify no reboot
    if start ~= nil and start <= time then
      res.updater = initialSpellArg.updater;
      res.duration = duration;
      res.texture = action:GetTexture();
      timer:StartTimer(res.name, res.duration, res.texture, res.target, res.count,
        res.tickFrequency, res.tickStart, res.updater);
      timer:SetTimeLeft(start+duration-time);
    end
  end
end


--[[
  The ManagerFrame class.
--]]

-- ManagerFrame
local ManagerFrame, ManagerFrameFactory, managerFrameFactory = factoryLib.Factory.CreateType();
managerFrameLib.ManagerFrame = ManagerFrame;
managerFrameLib.ManagerFrameFactory = ManagerFrameFactory;
managerFrameLib.managerFrameFactory = managerFrameFactory;

ManagerFrame.LEFT_DIRECTION = 1;
ManagerFrame.RIGHT_DIRECTION = 2;
ManagerFrame.DIRECTION_SIZE = 2;

ManagerFrame.actionBarSkin = {verticalGap=30, horizontalGap=30};
ManagerFrame.timerSkin = {visualRenderer = function (frame, barName) return tM2TimerRendererLib.tM2TimerRendererFactory:Create(frame); end, verticalGap=40, horizontalGap=100};
ManagerFrame.minimumGap = 35;

function ManagerFrame:SpreadState()
  local direction = self.backup.direction;
  local gap = max(self.actionBarSkin.verticalGap, self.timerSkin.verticalGap, self.minimumGap);
  if direction == self.LEFT_DIRECTION then
    for z, actionBarRenderer in ipairs(self.actionBarRenderer) do
      actionBarRenderer:SetPosition(self.frame, 5+self.actionBarSkin.horizontalGap/2, 
                                    5+gap*(z-.5));
      actionBarRenderer:SetAngle(math.pi + self.backup.angle);
    end
    for z, visualTimerRenderer in ipairs(self.visualTimerRenderer) do
      visualTimerRenderer:SetPosition(self.frame, 10+self.actionBarSkin.horizontalGap, 
                                      5+gap*(z-.5));
      visualTimerRenderer:SetDirection(1);  
    end
  elseif direction == self.RIGHT_DIRECTION then
    for z, actionBarRenderer in ipairs(self.actionBarRenderer) do
      actionBarRenderer:SetPosition(self.frame, 10+self.timerSkin.horizontalGap+self.actionBarSkin.horizontalGap/2,
                                    5+gap*(z-.5));
      actionBarRenderer:SetAngle(self.backup.angle);
    end
    for z, visualTimerRenderer in ipairs(self.visualTimerRenderer) do
      visualTimerRenderer:SetPosition(self.frame, 5, 
                                      5+gap*(z-.5));
      visualTimerRenderer:SetDirection(2);  
    end
  end
  local maxz = max(#self.actionBarRenderer, #self.visualTimerRenderer);
  self.frame:SetWidth(15+self.actionBarSkin.horizontalGap+self.timerSkin.horizontalGap);
  self.frame:SetHeight(10+maxz*gap);
end

function ManagerFrame:GetAngle(value)
  return self.backup.angle;
end

function ManagerFrame:SetAngle(value)
  self.backup.angle = value;
  self:SpreadState();
end

function ManagerFrame:SetDirection(direction)
  if direction >= 1 and direction <= self.DIRECTION_SIZE then
    self.backup.direction = direction;
  end
  self:SpreadState();
end

function ManagerFrame:SetActionBarRendererCreatorFunc(actionRendererCreatorFunc)
  self.actionBarRenderer = {};
  for _, actionBar in ipairs(self.actionBarList) do
    actionBarRenderer = actionRendererCreatorFunc();
    actionBar:SetActionBarRenderer(actionBarRenderer);
    table.insert(self.actionBarRenderer, actionBarRenderer);
  end
  self:SpreadState();
end

function ManagerFrame:GetDirection()
  return self.backup.direction;
end

function ManagerFrame:GetBackup()
  return self.backup;
end

function ManagerFrame:ToggleLocked()
  if self.locked then
    -- transparent colors
    self.frame:SetBackdropColor(0, 0, 0, 0);
    self.frame:SetBackdropBorderColor(.3, .3, .3, 0);
    self.frame:EnableMouse(false);
    self.locked = false;
  else
    self.frame:SetBackdropColor(0, 0, 0, 1);
    self.frame:SetBackdropBorderColor(.3, .3, .3, 1);
    self.frame:EnableMouse(true);
    self.locked = true;
  end
end

function ManagerFrame:IsVisible()
  return self.backup.visible;
end

function ManagerFrame:ToggleVisibility()
  if self.backup.visible then
    for _, visualTimerRenderer in ipairs(self.visualTimerRenderer) do
      visualTimerRenderer:SetEnabled(false);
    end
    self.frame:Hide();
    self.backup.visible = false;
  else
    self.frame:Show();
    for _, visualTimerRenderer in ipairs(self.visualTimerRenderer) do
      visualTimerRenderer:SetEnabled(true);
    end
    self.backup.visible = true;
  end    
end

function ManagerFrame:GetFirstSetButton(nextSetButton)
  return self.firstActionBar:GetSetButton();
end

function ManagerFrame:SetNextSetButton(nextSetButton)
  self.lastActionBar:SetNextSetButton(nextSetButton);
end

function ManagerFrame:IsLibSinkMessageVisible()
  return self.backup.libSinkEnabled;
end

function ManagerFrame:ToggleLibSinkMessageVisibility()
  if self.backup.libSinkEnabled then
    for _, messageTimerRenderer in ipairs(self.messageTimerRenderer) do
      messageTimerRenderer:SetEnabled(false);
    end
    self.backup.libSinkEnabled = false;
  else
    for _, messageTimerRenderer in ipairs(self.messageTimerRenderer) do
      messageTimerRenderer:SetEnabled(true);
    end
    self.backup.libSinkEnabled = true;
  end
end

--[[
  The ManagerFrameFactory class.
--]]

function ManagerFrameFactory:Create(frame, barNameList, actionBarConfMap, actionType, removeFuncMap, backup, messages, 
                                    actionRendererCreatorFunc, defaultLeft, defaultTop)
  local managerFrame = self:NewInstance();
  local barNameListSize = #barNameList;
  frame:ClearAllPoints();
  if backup ~= nil then
    if backup.left ~= nil and backup.top ~= nil then
      frame:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", backup.left, backup.top);
    else
      frame:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", defaultLeft or 0, defaultTop or 0);    
    end
    -- retro compatibility
    if backup.angle == nil then
      if backup.direction == 1 then
        backup.angle = -math.pi/8;
        backup.direction = ManagerFrame.LEFT_DIRECTION;
      elseif backup.direction == 2 then
        backup.angle = -math.pi/8;
        backup.direction = ManagerFrame.RIGHT_DIRECTION;
      elseif backup.direction == 3 then
        backup.angle = 0;
        backup.direction = ManagerFrame.LEFT_DIRECTION;
      elseif backup.direction == 4 then
        backup.angle = 0;
        backup.direction = ManagerFrame.RIGHT_DIRECTION;
      end
    end
    if backup.libSinkEnabled == nil then
      backup.libSinkEnabled = true;
    end
  else
    -- cannot initialize left and top (because these information may not be available)
    backup = {visible=true, direction=ManagerFrame.LEFT_DIRECTION, angle = -math.pi/8, libSinkEnabled = true};
    frame:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", defaultLeft or 0, defaultTop or 0);
  end
  frame:SetFrameStrata("LOW");
  frame:SetToplevel(true);
  frame:SetParent(UIParent);
  frame:SetWidth(270);
  frame:SetHeight(37*barNameListSize);
  frame:SetBackdrop(self.bckd);
  frame:SetBackdropColor(0 ,0 ,0 , 0);
  frame:SetBackdropBorderColor(.3, .3, .3, 0);
  frame:EnableMouse(false);
  frame:RegisterForDrag("LeftButton");
  frame:SetScript("OnDragStop",function (self)
    self:StopMovingOrSizing();
    self.managerFrame.backup.top = self:GetTop();
    self.managerFrame.backup.left = self:GetLeft(); 
  end);
  frame:SetScript("OnDragStart",function (self)
    self:StartMoving();
  end);
  managerFrame.frame = frame;
  frame.managerFrame = managerFrame;
  managerFrame.backup = backup;
  managerFrame.locked = false;
  
  managerFrame.actionBarList = {};
  managerFrame.actionBarRenderer = {};
  managerFrame.visualTimerRenderer = {};
  managerFrame.messageTimerRenderer = {};
  
  local previousActionBar = nil;
  managerFrame.firstActionBar = nil;
  
  -- foreach bar (ex: earth shield, elemental shield, ...)
  for z=1, barNameListSize do
    local barName = barNameList[z];
    local actionBarConfList = actionBarConfMap[barName];
    local actionList = {};
    local actionPropertyMap = {};
    local actionFuncMap = {};
    
    --[[
      SPECIAL for totem
    ]]
    local propertiesByTotemName = {};
    
    -- foreach action in bar (ex: earth elemental shield, water elemental shield, ...)
    for i=1,#actionBarConfList do
      local actionBarConf = actionBarConfList[i];
      if actionBarConf.action==nil or actionBarConf.action=="spell" then
        -- 1) create a spell action
        local spellAction = actionLib.spellActionFactory:Create(actionBarConf.name, actionBarConf.target);
        -- 2) the action property map will be given to timer when ACTION_PERFORMED occurs
        actionPropertyMap[spellAction] = actionBarConf;
        if actionType == "totem" then
          if actionBarConf.totemName == nil then
            actionBarConf.totemName = actionBarConf.name;
          end
          propertiesByTotemName[string.lower(actionBarConf.totemName)] = actionBarConf;
        elseif actionType == "buff" then
          if actionBarConf.buffName == nil then
            actionBarConf.buffName = actionBarConf.name;
          end        
        end
        -- the conf in action property map is not exactly what we want so we make a function which will be
        -- call after ACTION_PERFORMED occurs (see TimerLib.ActionPerformedTimer:ACTION_PERFORMED)
        if barName=="earthShield" then
          -- hook the timer duration (set it to 0 if you put earth shield on yourself)
          actionFuncMap[spellAction] = earthShieldTransformSpellArg;
        elseif actionType=="cooldown" then
          actionFuncMap[spellAction] = transformCooldownArg;        
        else
          actionFuncMap[spellAction] = transformSpellArg;
        end
        actionBarConf.action = spellAction;
        table.insert(actionList, spellAction);
      elseif actionBarConf.action=="item" then
        local itemAction = actionLib.itemActionFactory:Create(actionBarConf.name, actionBarConf.target, actionBarConf.equipItem);
        actionPropertyMap[itemAction] = actionBarConf;        
        actionFuncMap[itemAction] = transformItemArg;
        actionBarConf.action = itemAction;
        table.insert(actionList, itemAction);
      end
    end
        
    if actionType~="cooldown" then
      local actionBarRenderer = actionRendererCreatorFunc();
      table.insert(managerFrame.actionBarRenderer, actionBarRenderer);
      local backups = totemManagerData.actionBarBackup;
      if backups == nil then
        backups = {};
        totemManagerData.actionBarBackup = backups;
      end
            
      local actionBar = actionBarLib.actionBarFactory:Create(
         frame,
         barName,
         actionList,
         actionBarRenderer,
         backups[barName]
      );
      table.insert(managerFrame.actionBarList, actionBar);
      backups[barName] = actionBar:GetBackup();
      actionBar:SetSelectedCtrlClickFunc(removeFuncMap[barName]);
      if previousActionBar == nil then
        managerFrame.firstActionBar = actionBar;
      else
        previousActionBar:SetNextSetButton(actionBar:GetSetButton());
      end
      previousActionBar = actionBar;
    end

    local visualTimerRenderer = ManagerFrame.timerSkin.visualRenderer(frame, barName);
    table.insert(managerFrame.visualTimerRenderer, visualTimerRenderer);
    local rendererList = {visualTimerRenderer};
    local timerMessage = libSinkTimerRendererLib.libSinkTimerRendererFactory:Create(
      messages.started, messages.stopping, messages.stopped, messages.countAtZero
    );
    table.insert(managerFrame.messageTimerRenderer, timerMessage);
    timerMessage:SetEnabled(backup.libSinkEnabled);
    if actionType == "totem" then
       timerMessage:ActivatePour(true, true, true, true);
       table.insert(rendererList, timerMessage);       
    elseif actionType == "buff" then
       -- only expiring
       timerMessage:ActivatePour(false, true);
       table.insert(rendererList, timerMessage);
    end
    if actionType~="cooldown" then
    
      if actionType == "totem" then
        local f = actionLib.spellActionFactory:Create(totemManager.FIRST_CALL);
        local s = actionLib.spellActionFactory:Create(totemManager.SECOND_CALL);
        local t = actionLib.spellActionFactory:Create(totemManager.THIRD_CALL);
        table.insert(actionList, f);
        table.insert(actionList, s);
        table.insert(actionList, t);        
        actionFuncMap[f] = transformSummonArg;
        actionFuncMap[s] = transformSummonArg;
        actionFuncMap[t] = transformSummonArg;
        
        local sl;
        if barName == "airTotem" then
          sl = {slot=4};
        elseif barName == "waterTotem" then
          sl = {slot=3};
        elseif barName == "fireTotem" then
          sl = {slot=1};
        elseif barName == "earthTotem" then
          sl = {slot=2};
        end
        sl.propertiesByTotemName = propertiesByTotemName;
        actionPropertyMap[f] = sl;
        actionPropertyMap[s] = sl;
        actionPropertyMap[t] = sl;
      end
    
      timerLib.actionPerformedTimerFactory:Create(
        rendererList,
        actionList,
        actionPropertyMap,
        actionFuncMap
      );
    else
      timerLib.actionCooldownTimerFactory:Create(
        rendererList,
        actionList[1],
        actionPropertyMap[actionList[1]],
        actionFuncMap[actionList[1]]
      );    
    end
  end
  managerFrame.lastActionBar = previousActionBar;
  
  managerFrame:SpreadState();
  
  if backup.visible then
    frame:Show();
  else
    backup.visible = true;
    managerFrame:ToggleVisibility();
  end
  
  environment:Fire("SPELL_UPDATE_COOLDOWN");
  
  return managerFrame;
end

local bckd = {};
bckd.bgFile = "Interface/Tooltips/UI-Tooltip-Background";
bckd.edgeFile = "Interface/Tooltips/UI-Tooltip-Border";
bckd.tile = true;
bckd.tileSize = 16;
bckd.edgeSize = 16;
bckd.insets = {};
bckd.insets.left = 4;
bckd.insets.right = 4;
bckd.insets.top = 4;
bckd.insets.bottom = 4;
ManagerFrameFactory.bckd = bckd;

