--[[
    SquareActionBarRendererLib
    Copyright (C) 2008 Gael Lalire
    Contact: gael.lalire@laposte.net

    SquareActionBarRendererLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SquareActionBarRendererLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SquareActionBarRendererLib.  If not, see <http://www.gnu.org/licenses/>.
--]]

local squareActionBarRendererLib = LibStub:NewLibrary("SquareActionBarRendererLib-1.0", 1);
if not squareActionBarRendererLib then
  return;
end

-- begin import
local factoryLib = LibStub("FactoryLib-1.0");
local Factory = factoryLib.Factory;
local abstractActionBarRendererLib = LibStub("AbstractActionBarRendererLib-1.0");
local AbstractActionBarRendererFactory = abstractActionBarRendererLib.AbstractActionBarRendererFactory;
local AbstractActionBarRenderer = abstractActionBarRendererLib.AbstractActionBarRenderer;
local environmentLib = LibStub("EnvironmentLib-1.0");
local environment = environmentLib.environment;
-- end import

--[[
  The SquareActionBarRenderer class.
--]]

-- SquareActionBarRenderer extends ActionBarRenderer
local SquareActionBarRenderer, SquareActionBarRendererFactory, squareActionBarRendererFactory = Factory.CreateType(AbstractActionBarRendererFactory);
squareActionBarRendererLib.SquareActionBarRenderer = SquareActionBarRenderer;
squareActionBarRendererLib.SquareActionBarRendererFactory = SquareActionBarRendererFactory;
squareActionBarRendererLib.squareActionBarRendererFactory = squareActionBarRendererFactory;


function SquareActionBarRenderer:ApplySkin(secureUnitButton)
  local position = secureUnitButton.position;

  -- Show all graphical adds to the button

  secureUnitButton:SetWidth(self.buttonWidth);
  secureUnitButton:SetHeight(self.buttonHeigth);
  secureUnitButton:SetPoint("CENTER", self.point, "TOPLEFT", self.left+position*(self.escape*math.cos(self.angle)), -(self.top-position*(self.escape*math.sin(self.angle))));

  -- call parent
  AbstractActionBarRenderer.ApplySkin(self, secureUnitButton);  
end

function SquareActionBarRenderer:UnApplySkin(secureUnitButton)
  -- call parent
  AbstractActionBarRenderer.UnApplySkin(self, secureUnitButton);
end


function SquareActionBarRenderer:SetAngle(angle)
  self.angle = angle;
  local cosa = math.cos(angle);
  if cosa == 0 then
    self.escape = self.buttonHeight;
  elseif cosa==1 or cosa==-1 then
      self.escape = self.buttonWidth;
  else
    local sina = math.sin(angle);
    local cosEscape = math.abs(self.buttonWidth/cosa);
    local sinEscape = math.abs(self.buttonHeigth/sina);
    if cosEscape<sinEscape then
      self.escape = cosEscape;
    else
      self.escape = sinEscape;
    end
  end
  self:UpdateAllSkin();
end

function SquareActionBarRenderer:SetPosition(point, left, top)
  self.point = point;
  self.left = left;
  self.top = top;
  self:UpdateAllSkin();  
end

--[[
  The SquareActionBarRendererFactory class.
--]]

-- left and top indicate the position of the first button's center
function SquareActionBarRendererFactory:Create(tooltipLabels, buttonWidth, buttonHeigth)
  
  local squareActionBarRenderer = self:NewInstance(tooltipLabels, nil, nil, 0);
  
  squareActionBarRenderer.point = UIParent;
  squareActionBarRenderer.left = 0;
  squareActionBarRenderer.top = 0;
  squareActionBarRenderer.angle = 0;
  squareActionBarRenderer.escape = buttonWidth;
  squareActionBarRenderer.buttonWidth = buttonWidth;
  squareActionBarRenderer.buttonHeigth = buttonHeigth;
  
  return squareActionBarRenderer;
end
